import os
from typing import List, Sequence

import plotly.graph_objects as go
import torch
import xformers.ops as xops

from grouped_query_attention_pytorch.attention import (
    scaled_dot_product_gqa,
)
from grouped_query_attention_pytorch.utils.benchmark import BenchmarkResult, benchmark

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 if torch.cuda.is_available() else torch.float32

# Benchmarking parameters
NUM_GROUPS = [1, 4, 8, 16, 32, 64]
TOTAL_TOKENS = 8192
NUM_HEADS = 64
EMBED_DIM = 8
SEQ_LENGTH = 2048
SAVE_PATH = os.path.join("doc", "benchmark_attention.png")

BENCHMARK_SETTINGS_TEMPLATE = """
Benchmark settings:
    device: {device}
    dtype: {dtype}
    total_tokens: {total_tokens}
    seq_length: {seq_length}
    num_heads: {num_heads}
    embed_dim: {embed_dim}
"""


def main(
    num_groups: Sequence[int] = NUM_GROUPS,
    total_tokens: int = TOTAL_TOKENS,
    seq_length: int = SEQ_LENGTH,
    num_heads: int = NUM_HEADS,
    embed_dim: int = EMBED_DIM,
    device: torch.device = DEVICE,
    dtype: torch.dtype = DTYPE,
    save_path: str = SAVE_PATH,
):
    batch_size = total_tokens // seq_length
    q = torch.randn(
        batch_size, seq_length, num_heads, embed_dim, device=device, dtype=dtype
    )
    kv = torch.randn(
        batch_size, seq_length, num_heads, embed_dim, device=device, dtype=dtype
    )

    _ = scaled_dot_product_gqa(q, kv, kv)
    vanilla_result = benchmark(scaled_dot_product_gqa, q, kv, kv)
    print(f"Vanilla: {vanilla_result}")
    xformers_result = benchmark(xops.memory_efficient_attention, q, kv, kv)
    print(f"Flash Attn: {xformers_result}")

    grouped_times: List[BenchmarkResult] = []
    for g in num_groups:
        kv = torch.randn(
            batch_size, seq_length, g, embed_dim, dtype=dtype, device=device
        )
        grouped_result = benchmark(
            scaled_dot_product_gqa, q, kv, kv, force_grouped=True
        )
        grouped_times.append(grouped_result)
        print(f"Grouped (g={g}): {grouped_result}")

    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[vanilla_result.mean * 1000] * len(num_groups),
            mode="lines",
            line={"dash": "dash"},
            name="Vanilla MHA",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[r.mean * 1000 for r in grouped_times],
            mode="lines",
            name="GQA",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[grouped_times[0].mean * 1000] * len(num_groups),
            mode="lines",
            line={"dash": "dash"},
            name="MQA",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[xformers_result.mean * 1000] * len(num_groups),
            mode="lines",
            line={"dash": "dash"},
            name="Flash Attn (v1 - xformers)",
        )
    )
    fig.update_layout(
        title="Attention Benchmarks",
        xaxis_title="GQA Groups",
        yaxis_title="Runtime (ms)",
        # use log-scale for x-axis
        xaxis={"tickmode": "array", "tickvals": num_groups, "type": "log"},
        # place legend at center-left
        legend={"x": 0.1, "y": 0.5},
    )
    fig.write_image(save_path)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--num-groups",
        type=int,
        nargs="+",
        default=NUM_GROUPS,
        help="Sequence of GQA group sizes for benchmarking (default: [1, 4, 8, 16, 32, 64])",
    )
    parser.add_argument(
        "--total-tokens",
        type=int,
        default=TOTAL_TOKENS,
        help="Total number of tokens in the batch (default: 8192)",
    )
    parser.add_argument(
        "--seq-length",
        type=int,
        default=SEQ_LENGTH,
        help="Sequence length of the input (default: 2048)",
    )
    parser.add_argument(
        "--num-heads",
        type=int,
        default=NUM_HEADS,
        help="Number of attention heads (default: 64)",
    )
    parser.add_argument(
        "--embed-dim",
        type=int,
        default=EMBED_DIM,
        help="Embedding dimension of the input (default: 8)",
    )
    parser.add_argument(
        "--device",
        type=str,
        default=DEVICE,
        help="Device to run the benchmark on (default: 'cuda' if available else 'cpu')",
    )
    parser.add_argument(
        "--dtype",
        type=str,
        default=DTYPE,
        help="Data type to run the benchmark on (default: 'float16' if cuda is available else 'float32')",
    )
    parser.add_argument(
        "--save-path",
        type=str,
        default=SAVE_PATH,
        help="Path to save the benchmark plot (default: 'doc/benchmark_attention.png')",
    )
    args = parser.parse_args()

    print(
        BENCHMARK_SETTINGS_TEMPLATE.format(
            device=args.device,
            dtype=args.dtype,
            total_tokens=args.total_tokens,
            seq_length=args.seq_length,
            num_heads=args.num_heads,
            embed_dim=args.embed_dim,
        )
    )

    main(**vars(args))



import os
from typing import List, Sequence

import plotly.graph_objects as go
import torch
from torch import Tensor
from transformers import T5ForConditionalGeneration, T5Tokenizer

from grouped_query_attention_pytorch.t5 import convert_t5_to_gqa
from grouped_query_attention_pytorch.utils.benchmark import BenchmarkResult, benchmark

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 if torch.cuda.is_available() else torch.float32

# Benchmarking parameters
MODEL_NAME = "t5-3b"
NUM_GROUPS = [1, 2, 4, 8, 16, 32]
TOTAL_TOKENS = 4096
MODEL_MAX_LENGTH = 512
SAVE_PATH = os.path.join("doc", "benchmark_t5.png")

BENCHMARK_SETTINGS_TEMPLATE = """
Benchmark settings:
    device: {device}
    dtype: {dtype}
    model_name: {model_name}
    total_tokens: {total_tokens}
    model_max_length: {model_max_length}
"""
# NOTE: Text sample taken from the CNN/Daily Mail training set
INPUT_TEXT = """LONDON, England (Reuters) -- Harry Potter star Daniel Radcliffe gains
access to a reported £20 million ($41.1 million) fortune as he turns 18 on Monday, but
he insists the money won't cast a spell on him. Daniel Radcliffe as Harry Potter in
"Harry Potter and the Order of the Phoenix" To the disappointment of gossip
columnists around the world, the young actor says he has no plans to fritter his
cash away on fast cars, drink and celebrity parties. "I don't plan to be one of
those people who, as soon as they turn 18, suddenly buy themselves a massive sports
car collection or something similar," he told an Australian interviewer earlier this
month. "I don't think I'll be particularly extravagant. "The things I like buying
are things that cost about 10 pounds -- books and CDs and DVDs." At 18, Radcliffe
will be able to gamble in a casino, buy a drink in a pub or see the horror film
"Hostel: Part II," currently six places below his number one movie on the UK box
office chart. Details of how he'll mark his landmark birthday are under wraps.
His agent and publicist had no comment on his plans. "I'll definitely have some
sort of party," he said in an interview. "Hopefully none of you will be reading
about it." Radcliffe's earnings from the first five Potter films have been held
in a trust fund which he has not been able to touch. Despite his growing fame and
riches, the actor says he is keeping his feet firmly on the ground. "People are
always looking to say 'kid star goes off the rails,'" he told reporters last month.
"But I try very hard not to go that way because it would be too easy for them." His
latest outing as the boy wizard in "Harry Potter and the Order of the Phoenix" is
breaking records on both sides of the Atlantic and he will reprise the role in the
last two films. Watch I-Reporter give her review of Potter's latest » . There is life
beyond Potter, however. The Londoner has filmed a TV movie called "My Boy Jack," about
author Rudyard Kipling and his son, due for release later this year. He will also
appear in "December Boys," an Australian film about four boys who escape an orphanage.
Earlier this year, he made his stage debut playing a tortured teenager in Peter
Shaffer's "Equus." Meanwhile, he is braced for even closer media scrutiny now that
he's legally an adult: "I just think I'm going to be more sort of fair game," he told
Reuters. E-mail to a friend . Copyright 2007 Reuters. All rights reserved.This material
may not be published, broadcast, rewritten, or redistributed."""
TARGET_TEXT = """Harry Potter star Daniel Radcliffe gets £20M fortune as he turns
18 Monday. Young actor says he has no plans to fritter his cash away. Radcliffe's
earnings from first five Potter films have been held in trust fund."""


@torch.no_grad()
def forward_fn(model: T5ForConditionalGeneration, input_ids: Tensor, labels: Tensor):
    return model(input_ids=input_ids, labels=labels).loss


def main(
    model_name: str = MODEL_NAME,
    num_groups: Sequence[int] = NUM_GROUPS,
    total_tokens: int = TOTAL_TOKENS,
    model_max_length: int = MODEL_MAX_LENGTH,
    device: torch.device = DEVICE,
    dtype: torch.dtype = DTYPE,
    save_path: str = SAVE_PATH,
):
    print(f"Loading model and tokenizer for {model_name}...")
    t5: T5ForConditionalGeneration = T5ForConditionalGeneration.from_pretrained(
        model_name
    ).to(device=device, dtype=dtype)
    tokenizer = T5Tokenizer.from_pretrained(model_name, legacy=False)

    batch_size = total_tokens // model_max_length
    input_ids = tokenizer(
        [INPUT_TEXT] * batch_size,
        return_tensors="pt",
        max_length=model_max_length,
        truncation=True,
    ).input_ids.to(device=device)
    labels = tokenizer(
        [TARGET_TEXT] * batch_size,
        return_tensors="pt",
        max_length=model_max_length,
        truncation=True,
    ).input_ids.to(device=device)

    mha_result = benchmark(forward_fn, t5, input_ids, labels)
    print(f"MHA: {mha_result}")
    del t5
    torch.cuda.empty_cache()

    grouped_times: List[BenchmarkResult] = []
    for g in num_groups:
        print("Reloading model and converting to GQA...")
        t5 = T5ForConditionalGeneration.from_pretrained(model_name).to(
            device=device, dtype=dtype
        )
        gqa = convert_t5_to_gqa(t5, kv_heads=g, inplace=True)
        grouped_result = benchmark(forward_fn, gqa, input_ids, labels)
        grouped_times.append(grouped_result)
        print(f"Grouped (g={g}): {grouped_result}")

        del t5, gqa
        torch.cuda.empty_cache()

    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[mha_result.mean] * len(num_groups),
            mode="lines",
            line={"dash": "dash"},
            name="MHA",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[r.mean for r in grouped_times],
            mode="lines",
            name="GQA",
        )
    )
    fig.add_trace(
        go.Scatter(
            x=num_groups,
            y=[grouped_times[0].mean] * len(num_groups),
            mode="lines",
            line={"dash": "dash"},
            name="MQA",
        )
    )
    fig.update_layout(
        title="T5 Benchmarks",
        xaxis_title="GQA Groups",
        yaxis_title="Time per sample (s)",
        # use log-scale for x-axis
        xaxis={"tickmode": "array", "tickvals": num_groups, "type": "log"},
        # place legend at center-left
        legend={"x": 0.1, "y": 0.5},
    )
    fig.write_image(save_path)


if __name__ == "__main__":
    import argparse

    # NOTE: The original paper uses T5 v1.1 XL and XXL models.  When I load those
    # models through 'transformers' without applying GQA, I get nonsense outputs.
    # TODO: Figure out why this is happening.
    #   tokenizer = T5Tokenizer.from_pretrained("google/t5-v1_1-large", legacy=False)
    #   model = T5ForConditionalGeneration.from_pretrained("google/t5-v1_1-large")
    #
    # In the meantime, we can use the non-Google T5 models, which seem to work fine.
    # NOTE: Since the the original number of heads (n_heads) must be divisible by
    # 'kv_heads', there are only certain values of 'kv_heads' that we can use.
    # The following values of 'kv_heads' should be valid:
    #   - t5-small: 1, 2, 4, 8
    #   - t5-base: 1, 2, 3, 4, 6, 12
    #   - t5-large: 1, 2, 4, 8, 16
    #   - t5-3b: 1, 2, 4, 8, 16, 32  (DEFAULT)
    #   - t5-11b: 1, 2, 4, 8, 16, 32, 64  TODO: Check 11b values specifically

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model-name",
        type=str,
        default=MODEL_NAME,
        help=f"Name of the T5 model to benchmark (default: '{MODEL_NAME}')",
    )
    parser.add_argument(
        "--num-groups",
        type=int,
        nargs="+",
        default=NUM_GROUPS,
        help=f"Sequence of GQA group sizes for benchmarking (default: {NUM_GROUPS})",
    )
    parser.add_argument(
        "--total-tokens",
        type=int,
        default=TOTAL_TOKENS,
        help=f"Total number of tokens in the batch (default: {TOTAL_TOKENS})",
    )
    parser.add_argument(
        "--model-max-length",
        type=int,
        default=MODEL_MAX_LENGTH,
        help=f"Sequence length of the input (default: {MODEL_MAX_LENGTH})",
    )
    parser.add_argument(
        "--device",
        type=str,
        default=DEVICE,
        help=f"Device to run the benchmark on (default: {DEVICE})",
    )
    parser.add_argument(
        "--dtype",
        type=str,
        default=DTYPE,
        help=f"Data type to run the benchmark on (default: {DTYPE})",
    )
    parser.add_argument(
        "--save-path",
        type=str,
        default=SAVE_PATH,
        help=f"Path to save the benchmark plot (default: '{SAVE_PATH}')",
    )
    args = parser.parse_args()

    print(
        BENCHMARK_SETTINGS_TEMPLATE.format(
            device=args.device,
            dtype=args.dtype,
            model_name=args.model_name,
            total_tokens=args.total_tokens,
            model_max_length=args.model_max_length,
        )
    )

    main(**vars(args))



import pytest
import torch

from grouped_query_attention_pytorch.attention import (
    MultiheadGQA,
    scaled_dot_product_gqa,
)

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 if torch.cuda.is_available() else torch.float32
SEQ_LEN = 16


@pytest.mark.parametrize("embed_dim", [32])
@pytest.mark.parametrize("num_heads", [4, 8])
@pytest.mark.parametrize("kv_heads", [4, 8])
@pytest.mark.parametrize("is_causal", [True, False])
def test_grouped_scaled_dot_product_attention(
    embed_dim: int,
    num_heads: int,
    kv_heads: int,
    is_causal: bool,
):
    x = torch.randn(1, SEQ_LEN, num_heads, embed_dim, device=DEVICE, dtype=DTYPE)
    kv = torch.randn(1, SEQ_LEN, kv_heads, embed_dim, device=DEVICE, dtype=DTYPE)

    if kv_heads > num_heads:
        with pytest.raises(ValueError):
            scaled_dot_product_gqa(x, kv, kv, is_causal=is_causal)
        return

    out, attn_weights = scaled_dot_product_gqa(
        x, kv, kv, is_causal=is_causal, need_weights=True
    )
    assert out.size(0) == 1
    assert out.size(1) == SEQ_LEN
    assert out.size(2) == kv_heads
    assert out.size(3) == embed_dim
    assert attn_weights.size(0) == 1
    assert attn_weights.size(1) == SEQ_LEN
    assert attn_weights.size(2) == SEQ_LEN
    assert attn_weights.size(3) == kv_heads


@torch.no_grad()
@pytest.mark.parametrize("embed_dim", [64, 128])
@pytest.mark.parametrize("num_heads", [4, 8])
@pytest.mark.parametrize("kv_heads", [4, 8])
@pytest.mark.parametrize("is_causal", [True, False])
def test_multihead_gqa(
    embed_dim: int,
    num_heads: int,
    kv_heads: int,
    is_causal: bool,
):
    if kv_heads > num_heads:
        with pytest.raises(ValueError):
            MultiheadGQA(embed_dim, num_heads, kv_heads)
        return

    mhda = MultiheadGQA(embed_dim, num_heads, kv_heads, device=DEVICE, dtype=DTYPE)
    x = torch.randn(1, SEQ_LEN, embed_dim, device=DEVICE, dtype=DTYPE)

    out, _ = mhda(x, x, x, is_causal=is_causal)  # default: causal=False
    assert out.size(0) == 1
    assert out.size(1) == SEQ_LEN
    assert out.size(2) == embed_dim



from __future__ import annotations

from copy import deepcopy
from typing import TypeVar, overload

import torch
from einops import einsum, rearrange, repeat
from torch import nn
from transformers.models.t5.modeling_t5 import T5Attention


class T5GQA(nn.Module):
    def __init__(
        self,
        is_decoder: bool,
        d_model: int,
        key_value_proj_dim: int,
        n_heads: int,
        kv_heads: int,
        dropout: float,
        has_relative_attention_bias: bool,
        relative_attention_num_buckets: int,
        relative_attention_max_distance: int,
    ):
        super().__init__()
        if n_heads % kv_heads != 0:
            raise ValueError(
                f"n_heads ({n_heads}) must be divisible by kv_heads ({kv_heads})"
            )

        self.is_decoder = is_decoder
        self.d_model = d_model
        self.key_value_proj_dim = key_value_proj_dim
        self.n_heads = n_heads
        # TODO: Check if we need to store 'kv_heads' and 'inner_dim' as a properties
        self.kv_heads = kv_heads
        self.dropout = dropout
        # NOTE: Relative attention bias typically only used in the first layer
        # of a `T5Stack` module.
        self.has_relative_attention_bias = has_relative_attention_bias
        self.relative_attention_num_buckets = relative_attention_num_buckets
        self.relative_attention_max_distance = relative_attention_max_distance

        self.inner_dim = self.n_heads * self.key_value_proj_dim
        self.kv_dim = self.kv_heads * self.key_value_proj_dim

        # Mesh TensorFlow initialization to avoid scaling before softmax
        # self.q = nn.Linear(self.d_model, self.inner_dim, bias=False)
        # self.k = nn.Linear(self.d_model, self.kv_dim, bias=False)
        # self.v = nn.Linear(self.d_model, self.kv_dim, bias=False)
        # self.o = nn.Linear(self.kv_dim, self.d_model, bias=False)
        self.q = nn.Linear(self.d_model, self.inner_dim, bias=False)
        self.k = nn.Linear(self.d_model, self.inner_dim, bias=False)
        self.v = nn.Linear(self.d_model, self.inner_dim, bias=False)
        self.o = nn.Linear(self.inner_dim, self.d_model, bias=False)

        if self.has_relative_attention_bias:
            self.relative_attention_bias = nn.Embedding(
                self.relative_attention_num_buckets, self.n_heads
            )
        self.pruned_heads = set()  # type: ignore
        self.gradient_checkpointing = False

        self._relative_position_bucket = T5Attention._relative_position_bucket

    @classmethod
    def from_t5_attention(cls, t5: T5Attention, kv_heads: int) -> T5GQA:
        t5_gqa = T5GQA(
            is_decoder=t5.is_decoder,
            d_model=t5.d_model,
            key_value_proj_dim=t5.key_value_proj_dim,
            n_heads=t5.n_heads,
            kv_heads=kv_heads,
            dropout=t5.dropout,
            has_relative_attention_bias=t5.has_relative_attention_bias,
            relative_attention_num_buckets=t5.relative_attention_num_buckets,
            relative_attention_max_distance=t5.relative_attention_max_distance,
        )

        # Copy all of the weights verbatim from the original T5Attention module.
        # NOTE: In the T5 GQA implementation, all of the attention head aggregations
        # happen in the 'forward' method.  The weights themselves are not modified.
        t5_gqa.q.weight.data = t5.q.weight.data
        t5_gqa.k.weight.data = t5.k.weight.data
        t5_gqa.v.weight.data = t5.v.weight.data
        t5_gqa.o.weight.data = t5.o.weight.data
        if t5.has_relative_attention_bias:
            t5_gqa.relative_attention_bias.weight.data = (
                t5.relative_attention_bias.weight.data
            )

        return t5_gqa

    def forward(  # noqa: C901
        self,
        hidden_states,
        mask=None,
        key_value_states=None,
        position_bias=None,
        past_key_value=None,
        layer_head_mask=None,
        query_length=None,
        use_cache=False,
        output_attentions=False,
    ):
        """
        Self-attention (if key_value_states is None) or attention over source sentence (provided by key_value_states).
        """
        # Input is (batch_size, seq_length, dim)
        # Mask is (batch_size, key_length) (non-causal) or (batch_size, key_length, key_length)
        # past_key_value[0] is (batch_size, n_heads, q_len - 1, dim_per_head)
        batch_size, seq_length = hidden_states.shape[:2]

        real_seq_length = seq_length

        if past_key_value is not None:
            if len(past_key_value) != 2:
                raise ValueError(
                    f"past_key_value should have 2 past states: keys and values. Got { len(past_key_value)} past states"
                )
            real_seq_length += (
                past_key_value[0].shape[2] if query_length is None else query_length
            )

        key_length = (
            real_seq_length if key_value_states is None else key_value_states.shape[1]
        )

        def shape(states):
            """projection"""
            # NOTE: Changed from the original definition in T5Attention.
            sequence_length = states.shape[1]
            return states.view(
                batch_size, sequence_length, -1, self.key_value_proj_dim
            ).transpose(1, 2)

        def unshape(states):
            """reshape"""
            # NOTE: Changed from the original definition in T5Attention.
            sequence_length = states.shape[2]
            return (
                states.transpose(1, 2)
                .contiguous()
                .view(batch_size, sequence_length, -1)
            )

        def project(hidden_states, proj_layer, key_value_states, past_key_value):
            """projects hidden states correctly to key/query states"""
            if key_value_states is None:
                # self-attn
                # (batch_size, n_heads, seq_length, dim_per_head)
                hidden_states = shape(proj_layer(hidden_states))
            elif past_key_value is None:
                # cross-attn
                # (batch_size, n_heads, seq_length, dim_per_head)
                hidden_states = shape(proj_layer(key_value_states))

            if past_key_value is not None:
                if key_value_states is None:
                    # self-attn
                    # (batch_size, n_heads, key_length, dim_per_head)
                    hidden_states = torch.cat([past_key_value, hidden_states], dim=2)
                elif past_key_value.shape[2] != key_value_states.shape[1]:
                    # checking that the `sequence_length` of the `past_key_value` is the same as
                    # the provided `key_value_states` to support prefix tuning
                    # cross-attn
                    # (batch_size, n_heads, seq_length, dim_per_head)
                    hidden_states = shape(proj_layer(key_value_states))
                else:
                    # cross-attn
                    hidden_states = past_key_value
            return hidden_states

        # get query states: (batch_size, n_heads, seq_length, dim_per_head)
        grouped_queries = shape(self.q(hidden_states))
        # get key/value states
        key_states = project(
            hidden_states,
            self.k,
            key_value_states,
            past_key_value[0] if past_key_value is not None else None,
        )
        value_states = project(
            hidden_states,
            self.v,
            key_value_states,
            past_key_value[1] if past_key_value is not None else None,
        )

        # # compute scores
        # scores = torch.matmul(
        #     query_states, key_states.transpose(3, 2)
        # )  # equivalent of torch.einsum("bnqd,bnkd->bnqk", query_states, key_states), compatible with onnx op>9
        grouped_queries = rearrange(
            grouped_queries, "b (g h) n d -> b g h n d", h=self.kv_heads
        )
        grouped_keys = rearrange(
            key_states, "b (g h) s d -> b g h s d", h=self.kv_heads
        ).mean(dim=1)
        scores = einsum(grouped_queries, grouped_keys, "b g h n d, b h s d -> b h n s")

        if position_bias is None:
            if not self.has_relative_attention_bias:
                position_bias = torch.zeros(
                    # NOTE: This is different from the original in T5Attention!
                    # (1, self.n_heads, real_seq_length, key_length),
                    (1, self.kv_heads, real_seq_length, key_length),
                    device=scores.device,
                    dtype=scores.dtype,
                )
                if self.gradient_checkpointing and self.training:
                    position_bias.requires_grad = True
            else:
                position_bias = T5Attention.compute_bias(
                    self, real_seq_length, key_length, device=scores.device
                )

            # if key and values are already calculated
            # we want only the last query position bias
            if past_key_value is not None:
                position_bias = position_bias[:, :, -hidden_states.size(1) :, :]

            if mask is not None:
                # (batch_size, n_heads, seq_length, key_length)
                position_bias = position_bias + mask

        if self.pruned_heads:
            mask = torch.ones(position_bias.shape[1])
            mask[list(self.pruned_heads)] = 0
            position_bias_masked = position_bias[:, mask.bool()]
        else:
            position_bias_masked = position_bias

        # NOTE: This is different from the original in T5Attention!
        grouped_position_bias = rearrange(
            position_bias_masked, "b (g h) n s -> b g h n s", h=self.kv_heads
        ).mean(dim=1)

        scores += grouped_position_bias
        # attn_weights: (batch_size, kv_heads, seq_length, key_length)
        attn_weights = nn.functional.softmax(scores.float(), dim=-1).type_as(scores)
        attn_weights = nn.functional.dropout(
            attn_weights, p=self.dropout, training=self.training
        )

        # Mask heads if we want to
        if layer_head_mask is not None:
            attn_weights = attn_weights * layer_head_mask

        # NOTE: This is different from the original in T5Attention!
        # attn_output = unshape(torch.matmul(attn_weights, value_states))
        grouped_values = rearrange(
            value_states, "b (g h) s d -> b g h s d", h=self.kv_heads
        ).mean(dim=1)
        attn_output = unshape(torch.matmul(attn_weights, grouped_values))
        attn_output = repeat(
            attn_output, "b s d -> b s (g d)", g=(self.n_heads // self.kv_heads)
        )
        attn_output = self.o(attn_output)

        present_key_value_state = (
            (key_states, value_states) if (self.is_decoder and use_cache) else None
        )
        outputs = (attn_output,) + (present_key_value_state,) + (position_bias,)

        if output_attentions:
            outputs = outputs + (attn_weights,)  # type: ignore
        return outputs


ModuleType = TypeVar("ModuleType", bound=nn.Module)


@overload
def convert_t5_to_gqa(
    module: ModuleType, kv_heads: int, inplace: bool = False
) -> ModuleType:
    ...


@overload
def convert_t5_to_gqa(
    module: T5Attention, kv_heads: int, inplace: bool = False
) -> T5GQA:
    ...


def convert_t5_to_gqa(module, kv_heads: int, inplace: bool = False):
    if isinstance(module, T5Attention):
        return T5GQA.from_t5_attention(module, kv_heads=kv_heads)

    out = module if inplace else deepcopy(module)
    for name, child in out.named_children():
        out._modules[name] = convert_t5_to_gqa(child, kv_heads=kv_heads, inplace=True)
    return out


if __name__ == "__main__":
    from transformers import T5ForConditionalGeneration, T5Tokenizer

    # NOTE: The original paper uses T5 v1.1 XL and XXL models.  When I load those
    # models through 'transformers' without applying GQA, I get nonsense outputs.
    # TODO: Figure out why this is happening.
    #   tokenizer = T5Tokenizer.from_pretrained("google/t5-v1_1-large", legacy=False)
    #   model = T5ForConditionalGeneration.from_pretrained("google/t5-v1_1-large")
    #
    # In the meantime, we can use the non-Google T5 models, which seem to work fine.
    # NOTE: Since the the original number of heads (n_heads) must be divisible by
    # 'kv_heads', there are only certain values of 'kv_heads' that we can use.
    # To the best of my knowledge, the following values of 'kv_heads' are valid:
    #   - t5-small: 1, 2, 4, 8
    #   - t5-base: 1, 2, 3, 4, 6, 12
    #   - t5-large: 1, 2, 4, 8, 16
    #   - t5-3b: 1, 2, 4, 8, 16, 32
    #   - t5-11b: 1, 2, 4, 8, 16, 32, 64  TODO: Check 11b values specifically

    tokenizer = T5Tokenizer.from_pretrained(
        "t5-base", legacy=False, model_max_length=512
    )
    t5: T5ForConditionalGeneration = T5ForConditionalGeneration.from_pretrained(
        "t5-base"
    )
    gqa = convert_t5_to_gqa(t5, kv_heads=6)

    input_ids = tokenizer(
        "translate English to German: The house is wonderful.", return_tensors="pt"
    ).input_ids
    y2 = gqa.generate(input_ids, max_new_tokens=25)
    text = tokenizer.batch_decode(y2[0], skip_special_tokens=True)
    print(text)
    # The correct answer is:  ['<pad>', 'Das', 'Haus', 'ist', 'wunderbar', '.', '</s>']
    # NOTE: The original T5 model produces this answer, and so does GQA when we use
    # the maximum number of heads -- effectively equivalent to the original T5 model
    # with MHA.  The text quickly degrades as we reduce the number of heads.

    labels = tokenizer("Das Haus ist wunderbar.", return_tensors="pt").input_ids
    loss = gqa(input_ids=input_ids, labels=labels).loss
    print(f"Loss: {loss}")
    # NOTE: As above, the loss quickly degrades (increases) as we reduce the number
    # of GQA heads.



from math import log
from typing import Callable, Optional, Union

import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch.nn.modules.transformer import _get_activation_fn
from torchscale.component.xpos_relative_position import XPOS

from grouped_query_attention_pytorch.attention import MultiheadGQA


class GQATransformerEncoderLayer(nn.Module):
    # NOTE: Mostly pulled from 'nn.TransformerEncoderLayer', but with changes:
    #   - use sub-LayerNorm like in MAGNETO. See: https://arxiv.org/abs/2210.06423
    #   - use MultiheadGQA instead of MultiheadAttention

    def __init__(
        self,
        d_model: int,
        nhead: int,
        kv_heads: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        gamma_init: float = 1.0,
        device: Optional[Union[torch.device, str]] = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        super().__init__()
        # Legacy string support for activation function.
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)

        self.activation = activation
        self.gamma_init = gamma_init

        self.dropout = nn.Dropout(dropout)
        # Self-attention block
        self.norm1 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.self_attn = MultiheadGQA(  # type: ignore
            embed_dim=d_model,
            query_heads=nhead,
            kv_heads=kv_heads,
            dropout=dropout,
            layer_norm=True,
            layer_norm_eps=layer_norm_eps,
            gamma_init=gamma_init,
            device=device,
            dtype=dtype,
        )
        # Feedforward block
        self.norm2 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.linear1 = nn.Linear(d_model, dim_feedforward, device=device, dtype=dtype)
        self.norm3 = nn.LayerNorm(
            dim_feedforward, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.linear2 = nn.Linear(dim_feedforward, d_model, device=device, dtype=dtype)

        self._reset_parameters()

    def _reset_parameters(self):
        # NOTE: We follow the initialization strategy from MAGNETO.  See:
        # https://arxiv.org/pdf/2210.06423.pdf, Fig. 2
        # The 'MultiheadGQA' module uses ths same initialization,
        # so we just need to worry about the 'Linear' modules here.
        nn.init.xavier_normal_(self.linear1.weight, gain=self.gamma_init)
        nn.init.constant_(self.linear1.bias, 0)
        nn.init.xavier_normal_(self.linear2.weight, gain=self.gamma_init)
        nn.init.constant_(self.linear2.bias, 0)

    def _self_attention_block(self, x: Tensor, is_causal: bool = False) -> Tensor:
        x = self.norm1(x)
        x, _ = self.self_attn(x, x, x, is_causal=is_causal)
        x = self.dropout(x)
        return x

    def _feedforward_block(self, x: Tensor) -> Tensor:
        x = self.norm2(x)
        x = self.activation(self.linear1(x))
        x = self.dropout(x)
        x = self.norm3(x)
        x = self.linear2(x)
        x = self.dropout(x)
        return x

    def forward(self, src: Tensor, is_causal: bool = False) -> Tensor:
        x = src
        x = x + self._self_attention_block(x, is_causal=is_causal)
        x = x + self._feedforward_block(x)
        return x


class GQATransformerDecoderLayer(nn.Module):
    # NOTE: Mostly pulled from 'nn.TransformerDecoderLayer', but with changes:
    #   - use sub-LayerNorm like in MAGNETO. See: https://arxiv.org/abs/2210.06423
    #   - use MultiheadGQA instead of MultiheadAttention

    def __init__(
        self,
        d_model: int,
        nhead: int,
        kv_heads: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        gamma_init: float = 1.0,
        device: Optional[Union[torch.device, str]] = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        super().__init__()
        # Legacy string support for activation function.
        if isinstance(activation, str):
            activation = _get_activation_fn(activation)

        self.activation = activation
        self.gamma_init = gamma_init

        self.dropout = nn.Dropout(dropout)
        # Self-attention block
        self.norm1 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.self_attn = MultiheadGQA(  # type: ignore
            embed_dim=d_model,
            query_heads=nhead,
            kv_heads=kv_heads,
            dropout=dropout,
            layer_norm=False,
            gamma_init=gamma_init,
            device=device,
            dtype=dtype,
        )
        # Multi-head attention block
        self.norm2 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.multihead_attn = MultiheadGQA(  # type: ignore
            embed_dim=d_model,
            query_heads=nhead,
            kv_heads=kv_heads,
            dropout=dropout,
            layer_norm_eps=layer_norm_eps,
            gamma_init=gamma_init,
            device=device,
            dtype=dtype,
        )
        # Feedforward block
        self.norm3 = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.linear1 = nn.Linear(d_model, dim_feedforward, device=device, dtype=dtype)
        self.norm4 = nn.LayerNorm(
            dim_feedforward, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.linear2 = nn.Linear(dim_feedforward, d_model, device=device, dtype=dtype)

        self._reset_parameters()

    def _reset_parameters(self):
        # NOTE: We follow the initialization strategy from MAGNETO.  See:
        # https://arxiv.org/pdf/2210.06423.pdf, Fig. 2
        # The 'MultiheadGQA' module uses ths same initialization,
        # so we just need to worry about the 'Linear' modules here.
        nn.init.xavier_normal_(self.linear1.weight, gain=self.gamma_init)
        nn.init.constant_(self.linear1.bias, 0)
        nn.init.xavier_normal_(self.linear2.weight, gain=self.gamma_init)
        nn.init.constant_(self.linear2.bias, 0)

    def _self_attention_block(self, x: Tensor, is_causal: bool = False) -> Tensor:
        x = self.norm1(x)
        x, _ = self.self_attn(x, x, x, is_causal=is_causal)
        x = self.dropout(x)
        return x

    def _multihead_attention_block(
        self, x: Tensor, memory: Tensor, is_causal: bool = False
    ) -> Tensor:
        x = self.norm2(x)
        x, _ = self.multihead_attn(x, memory, memory, is_causal=is_causal)
        x = self.dropout(x)
        return x

    def _feedforward_block(self, x: Tensor) -> Tensor:
        x = self.norm3(x)
        x = self.activation(self.linear1(x))
        x = self.dropout(x)
        x = self.norm4(x)
        x = self.linear2(x)
        x = self.dropout(x)
        return x

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
    ) -> Tensor:
        x = tgt
        x = x + self._self_attention_block(x, is_causal=tgt_is_causal)
        x = x + self._multihead_attention_block(x, memory, is_causal=memory_is_causal)
        x = x + self._feedforward_block(x)
        return x


class GQATransformer(nn.Module):
    def __init__(
        self,
        d_model: int = 512,
        nhead: int = 8,
        kv_heads: int = 4,
        num_encoder_layers: int = 6,
        num_decoder_layers: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        super().__init__()
        # The 'gamma_init' parameters are different for the encoder and decoder,
        # and depend on the number of encoder/decoder layers.  See MAGNETO paper:
        # https://arxiv.org/pdf/2210.06423.pdf, Figure 2
        encoder_gamma_init = (
            log(3 * num_decoder_layers) * log(2 * num_encoder_layers) / 3
        ) ** 0.5
        decoder_gamma_init = log(3 * num_decoder_layers) ** 0.5

        self.encoder = nn.TransformerEncoder(
            encoder_layer=GQATransformerEncoderLayer(
                d_model=d_model,
                nhead=nhead,
                kv_heads=kv_heads,
                dim_feedforward=dim_feedforward,
                dropout=dropout,
                activation=activation,
                layer_norm_eps=layer_norm_eps,
                gamma_init=encoder_gamma_init,
                device=device,
                dtype=dtype,
            ),
            num_layers=num_encoder_layers,
            mask_check=False,
            enable_nested_tensor=False,
        )
        self.decoder = nn.TransformerDecoder(
            decoder_layer=GQATransformerDecoderLayer(
                d_model=d_model,
                nhead=nhead,
                kv_heads=kv_heads,
                dim_feedforward=dim_feedforward,
                dropout=dropout,
                activation=activation,
                layer_norm_eps=layer_norm_eps,
                gamma_init=decoder_gamma_init,
                device=device,
                dtype=dtype,
            ),
            num_layers=num_decoder_layers,
        )

    def forward(self, x: Tensor, is_causal: bool = True) -> Tensor:
        """
        Input shape: (batch_size, seq_len, d_model)
        Output shape: (batch_size, seq_len, d_model)

        NOTE: Assume that 'is_causal' applies to both the encoder and decoder.
        This is the case for language modeling, but maybe not for other tasks.
        """
        tgt = x
        for layer in self.encoder.layers:
            x = layer(x, is_causal=is_causal)
        if self.encoder.norm is not None:
            x = self.encoder.norm(x)

        mem = x
        for layer in self.decoder.layers:
            tgt = layer(tgt, mem, memory_is_causal=is_causal, tgt_is_causal=is_causal)
        if self.decoder.norm is not None:
            tgt = self.decoder.norm(tgt)

        return tgt


class GQATransformerLM(nn.Module):
    def __init__(
        self,
        num_tokens: int,  # (required) usually obtained from the tokenizer
        d_model: int = 512,
        nhead: int = 8,
        kv_heads: int = 4,
        num_encoder_layers: int = 6,
        num_decoder_layers: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Union[str, Callable[[Tensor], Tensor]] = F.relu,
        layer_norm_eps: float = 1e-5,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        super().__init__()
        self.token_embedding = nn.Embedding(
            num_tokens, d_model, device=device, dtype=dtype
        )
        # TODO: Add support for other positional encodings?  I use XPOS, which is the
        # "latest and greatest" at the time of writing.  In principle, we could swap
        # it out for any other encoding, and remove the 'torchscale' dependency for this
        # repo, which is only used for XPOS.
        self.pos_embedding = XPOS(d_model).to(device=device, dtype=dtype)
        self.transformer = GQATransformer(
            d_model=d_model,
            nhead=nhead,
            kv_heads=kv_heads,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=activation,
            layer_norm_eps=layer_norm_eps,
            device=device,
            dtype=dtype,
        )
        self.norm = nn.LayerNorm(
            d_model, eps=layer_norm_eps, device=device, dtype=dtype
        )
        self.out = nn.Linear(d_model, num_tokens, device=device, dtype=dtype)

    def _reset_parameters(self):
        nn.init.kaiming_normal_(self.out.weight)
        nn.init.constant_(self.out.bias, 0)

    def forward(self, x: Tensor, is_causal: bool = True) -> Tensor:
        x = self.token_embedding(x)
        x = x + self.pos_embedding(x)
        x = self.transformer(x, is_causal=is_causal)
        x = self.norm(x)
        return self.out(x)


if __name__ == "__main__":
    num_tokens = 2048
    device = torch.device("cuda")
    dtype = torch.float16

    x = torch.randint(0, num_tokens - 1, size=(2, 512), device=device)
    model = GQATransformerLM(num_tokens=num_tokens, device=device, dtype=dtype)

    with torch.no_grad():
        out = model(x)
    print(out.shape)



import pytest
import torch
from transformers import T5ForConditionalGeneration, T5Tokenizer
from transformers.models.t5.modeling_t5 import T5Attention, T5LayerFF

from grouped_query_attention_pytorch.t5 import T5GQA, convert_t5_to_gqa

MODEL_NAME = "t5-small"
DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 if torch.cuda.is_available() else torch.float32
SEQ_LEN = 16


# Test all of the valid kv_heads values for 't5-small
@pytest.mark.parametrize("kv_heads", [1, 2, 4, 8])
def test_convert_t5_to_gqa(kv_heads: int):
    t5: T5ForConditionalGeneration = T5ForConditionalGeneration.from_pretrained(
        MODEL_NAME
    ).to(device=DEVICE, dtype=DTYPE)

    # Just to establish that the T5 model is as expected.  Check that all of the
    # known attention layers are of type 'T5Attention'.
    for block in t5.encoder.block:
        for layer in block.layer:
            if hasattr(layer, "SelfAttention"):
                assert isinstance(layer.SelfAttention, T5Attention)
            else:
                assert isinstance(layer, T5LayerFF)
    for block in t5.decoder.block:
        for layer in block.layer:
            if hasattr(layer, "SelfAttention"):
                assert isinstance(layer.SelfAttention, T5Attention)
            elif hasattr(layer, "EncDecAttention"):
                assert isinstance(layer.EncDecAttention, T5Attention)
            else:
                assert isinstance(layer, T5LayerFF)

    gqa = convert_t5_to_gqa(t5, kv_heads=kv_heads, inplace=True)
    # After conversion, check that all of the attention layers above have been
    # replaced with 'T5GQA' layers.
    for block in gqa.encoder.block:
        for layer in block.layer:
            if hasattr(layer, "SelfAttention"):
                assert isinstance(layer.SelfAttention, T5GQA)
            else:
                assert isinstance(layer, T5LayerFF)
    for block in t5.decoder.block:
        for layer in block.layer:
            if hasattr(layer, "SelfAttention"):
                assert isinstance(layer.SelfAttention, T5GQA)
            elif hasattr(layer, "EncDecAttention"):
                assert isinstance(layer.EncDecAttention, T5GQA)
            else:
                assert isinstance(layer, T5LayerFF)

    # Check that we can pass inputs/targets through the modified model, and that
    # the is returns a loss that is a scalar Tensor.
    tokenizer = T5Tokenizer.from_pretrained(MODEL_NAME, legacy=False)
    inputs = tokenizer(
        "translate English to German: The house is wonderful.", return_tensors="pt"
    )
    targets = tokenizer("Das Haus ist wunderbar.", return_tensors="pt")
    with torch.no_grad():
        loss = gqa(
            inputs.input_ids.to(DEVICE),
            labels=targets.input_ids.to(DEVICE),
        ).loss
    assert isinstance(loss, torch.Tensor)
    assert loss.size() == ()

    # Check that we can generate outputs, using the usual 'generate' method.
    out = gqa.generate(inputs.input_ids.to(DEVICE), max_new_tokens=10)
    assert isinstance(out, torch.Tensor)



from typing import Callable, Union

import pytest
import torch

from grouped_query_attention_pytorch.transformer import GQATransformer, GQATransformerLM

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DTYPE = torch.float16 if torch.cuda.is_available() else torch.float32
SEQ_LEN = 16


@pytest.mark.parametrize("d_model", [128])
@pytest.mark.parametrize("nhead", [4, 8])
@pytest.mark.parametrize("kv_heads", [2, 4])
@pytest.mark.parametrize("num_layers", [1, 2])
@pytest.mark.parametrize("dim_feedforward", [64])
@pytest.mark.parametrize("dropout", [0.0])
@pytest.mark.parametrize("activation", ["relu", "gelu"])
@pytest.mark.parametrize("is_causal", [True, False])
def test_gqa_transformer(
    d_model: int,
    nhead: int,
    kv_heads: int,
    num_layers: int,
    dim_feedforward: int,
    dropout: float,
    activation: Union[str, Callable],
    is_causal: bool,
):
    net = GQATransformer(
        d_model=d_model,
        nhead=nhead,
        kv_heads=kv_heads,
        num_encoder_layers=num_layers,
        num_decoder_layers=num_layers,
        dim_feedforward=dim_feedforward,
        dropout=dropout,
        activation=activation,
        device=DEVICE,
        dtype=DTYPE,
    )
    x = torch.randn(1, SEQ_LEN, d_model, device=DEVICE, dtype=DTYPE)
    with torch.no_grad():
        y = net.forward(x, is_causal=is_causal)
    assert y.size(0) == 1
    assert y.size(1) == SEQ_LEN
    assert y.size(2) == d_model


@pytest.mark.parametrize("num_tokens", [100])
@pytest.mark.parametrize("d_model", [128])
@pytest.mark.parametrize("nhead", [4, 8])
@pytest.mark.parametrize("kv_heads", [2, 4])
@pytest.mark.parametrize("num_layers", [1, 2])
@pytest.mark.parametrize("dim_feedforward", [64])
@pytest.mark.parametrize("dropout", [0.0])
@pytest.mark.parametrize("activation", ["relu", "gelu"])
@pytest.mark.parametrize("is_causal", [True, False])
def test_gqa_transformer_lm(
    num_tokens: int,
    d_model: int,
    nhead: int,
    kv_heads: int,
    num_layers: int,
    dim_feedforward: int,
    dropout: float,
    activation: Union[str, Callable],
    is_causal: bool,
):
    net = GQATransformerLM(
        num_tokens=num_tokens,
        d_model=d_model,
        nhead=nhead,
        kv_heads=kv_heads,
        num_encoder_layers=num_layers,
        num_decoder_layers=num_layers,
        dim_feedforward=dim_feedforward,
        dropout=dropout,
        activation=activation,
        device=DEVICE,
        dtype=DTYPE,
    )
    x = torch.randint(0, num_tokens, (1, SEQ_LEN), device=DEVICE, dtype=torch.long)
    with torch.no_grad():
        y = net.forward(x, is_causal=is_causal)
    assert y.size(0) == 1
    assert y.size(1) == SEQ_LEN
    assert y.size(2) == num_tokens
